/* Copyright (c) 2007-2012 Sam Trenholme
 *
 * TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

#include "DwMararc.h"
#include "DwStr.h"
#include "DwSys.h"
#include "DwStr_functions.h"
#include <stdint.h>
#include <stdio.h>

/* Number of string parameters in the mararc file */
#define KEY_S_COUNT 6
/* Number of dictionary parameters in the mararc file */
#define KEY_D_COUNT 1
/* Number of numeric parameters in the mararc file */
#define KEY_N_COUNT 17

dwm_fs fsm[DWM_MAX_STATES + 1]; /* Finite state machine */
dw_str *key_s[KEY_S_COUNT + 1]; /* All of the string dwood2rc parameters */
dw_str *key_d[KEY_D_COUNT + 1]; /* The dictionary dwood2rc parameters */
int32_t key_n[KEY_N_COUNT + 1]; /* The numeric dwood2rc parameters */

char *key_s_names[KEY_S_COUNT + 1] = {
        "bind_address", /* IP addresses to bind to */
        "ipv4_bind_addresses", /* IP Addresses to bind to (newer name) */
        "chroot_dir", /* Where we run Deadwood from */
        "recursive_acl", /* IPs that we allow recursive queries from */
        "random_seed_file", /* File with seed/key for random number
                             * generator */
        "cache_file", /* File with a copy of Deadwood's cache */
        0 };

char *key_d_names[KEY_D_COUNT + 1] = {
        "upstream_servers", /* Where we get our DNS info from */
        0 };

char *key_n_names[KEY_N_COUNT + 1] = {
        "maxprocs", /* The maximum number of outstanding queries we can
                     * have */
        "timeout_seconds", /* How long we wait for a reply from the remote
                            * server */
        "dns_port", /* The port we bind to */
        "upstream_port", /* The port we connect to when connecting upstream */
        "handle_overload", /* Reply when overloaded */
	"handle_noreply", /* Reply when no reply from upstream */
        "recurse_min_bind_port", /* The lowest numbered port deadwood will
                                  * bind to */
        "recurse_number_ports", /* The number of ports deadwood is allowed
                                 * to bind to */
        "hash_magic_number", /* A large 31-bit prime number that the hash
                              * compression function uses */
        "maximum_cache_elements", /* Maximum number of elements in cache */
	"maradns_uid", /* Numeric User ID Deadwood runs as */
	"maradns_gid", /* Numeric Group ID Deadwood runs as */
	"resurrections", /* Whether to look up expired records if we can't
                          * connect to upstream */
	"num_retries", /* Number of times we try to connect to an upstream 
                        * server before giving up */
	"verbose_level", /* How verbose our logging should be */
	"deliver_all", /* Deliver non-cachable replies */
	"max_ttl", /* Maximum allowed TTL */
        0 };

char *fsm_desc=dwm_machine;

int dwm_linenum = 0;

/* Set the new fsm state for a given pattern seen in a given fsm state
 * Input: c: Pointer to string with FSM description (pointing to the
 *           character we are currently parsing)
 *        d: The action number for this state that we are processing
 *        state: The state we are processing
 * Output: A revised pointer to the FSM description */
char *dwm_set_newstate(char *c, int d, int32_t state) {
        if(*c < 'a' || *c > 'z') {
                return 0;
        }
        /* If the state starts with x, it's an "expanded" state, with a
         * two-letter name like "xa", "xc", "xp", or "xz" */
        if(*c == 'x') {
                c++;
                if(*c < 'a' || *c > 'z') {
                        return 0;
                }
                fsm[state].newstate[d] = *c - 'a' + 26;
        } else {
                fsm[state].newstate[d] = *c - 'a';
        }
        c++;
        return c;
}

/* Set the action for a given pattern in a given fsm state
 * Input: c: Pointer to string with FSM description (pointing to the
 *           character we are currently parsing)
 *        max: How far we are in the current line (bounds checker)
 *        d: The action number for this state that we are processing
 *        state: The state we are processing
 * Output: A revised pointer to the FSM description */
char *dwm_set_action(char *c, int max, int d, int32_t state) {
        int b = 0;
        if(*c == ';') {
                fsm[state].action[d] = 10; /* Terminate with success */
                fsm[state].newstate[d] = 0;
                c++;
                max++;
                if(max > 52) {
                        return 0;
                }
                return c;
        } else {
                b = *c - '0';
                if(b < 1 || b > 9) {
                        return 0;
                }
                fsm[state].action[d] = b;
        }
        c++;
        max++;
        if(max >= 52) {
                return 0;
        }
        return dwm_set_newstate(c,d,state);
}

/* Convert a one-character FSM class in to a number we put in to
 * the tokenized FSM machine */
char dwm_pattern_process(int b) {
        /* We don't allow literal hashes or quotes in the FSM definition */
        if(b == '#' || b == '"') {
                return 0;
        }
        /* H becomes a hash symbol (so a FSM definition can have comments) */
        if(b == 'H') {
                b = '#';
        }
        /* Q becomes a " symbol (so we don't have ugly escape quoted sequences
         * in the FSM definition) */
        if(b == 'Q') {
                b = '"';
        }
        /* Make letter shortcuts non-printable ASCII control characters */
        if(b >= 'A' && b <= 'Z') {
                b = b - '@'; /* '@' is 'A' - 1 in ASCII */
        }
        return b;
}

/* Set a single state in the finite state machine */
char *dwm_set_fsm(int32_t state, char *c, int max) {
        int b = 0 , d = 0;

        for(; max<52; max++) {

                /* First letter: pattern */
                b = *c;
                if(b < '!' || b > '~') {
                        return 0;
                }
                fsm[state].pattern[d] = dwm_pattern_process(b);

                /* Second latter: action -or- next state */
                c++;
                max++;
                if(*c >= 'a' && *c <= 'z') {
                        fsm[state].action[d] = 0;
                        c = dwm_set_newstate(c,d,state);
                } else {
                        c = dwm_set_action(c,max,d,state);
                }

                /* Sanity check */
                if(c == 0) {
                        return 0;
                }

                /* Advance past whitespace between patterns/actions */
                while(*c == ' ') {
                        c++;
                        max++;
                        if(max >= 52) {
                                return 0;
                        }
                }

                if(*c == '\r' || *c == '\n') { /* Newline ends a state */
                        return c;
                }
                d++;
                if(d >= DWM_MAX_PATTERNS - 1) { /* Bounds checking */
                        return 0;
                }
        }
        return 0;
}

/* Tokenize a single line describing the finite state machine */
char *dwm_tokenize_line(char *c) {
        int b = 0;
        int max = 0;

        /* The first character of a line is the state we are describing */
        if(*c < 'a' || *c > 'z') {
                return 0;
        }
        /* If the first character happens to be an 'x', the state description
         * is two letters long in a form like "xa", "xb", and so on */
        if(*c == 'x') {
                c++;
                max++;
                if(*c < 'a' || *c > 'z') {
                        return 0;
                }
                b = *c - 'a' + 26;
        } else {
                b = *c - 'a';
        }
        if(b < 0 || b > DWM_MAX_STATES - 1) {
                return 0;
        }
        /* OK, now move forward to see the first pattern description */
        c++;
        if(*c != ' ') {
                return 0;
        }
        while(*c == ' ' && max < 52) {
                c++;
                max++;
        }
        if(max >= 52) {
                return 0;
        }
        /* dwm_set_fsm sets how we react to various patterns when in this
         * state */
        c = dwm_set_fsm(b,c,max);
        if(c == 0) {
                return 0;
        }
        if(*c != '\r' && *c != '\n') {
                return 0;
        }
        c++;
        return c;
}

/* Initialize the finite state machine based on the vales set in the
 * dwm_machine constant (which is set in DwMararc.h) */
void dwm_init_fsm() {
        char *c = 0;
        int a = 0;
        int b = 0;

        /* First, we zero it out */
        for(a=0;a<DWM_MAX_STATES;a++) {
                for(b=0;b<DWM_MAX_PATTERNS;b++) {
                        fsm[a].pattern[b] = 0;
                        fsm[a].action[b] = 0;
                        fsm[a].newstate[b] = 0;
                }
        }

        /* Now, we look at the defined fsm and tokenize it */
        c=fsm_desc;
        b=0;
        for(a=0;a<DWM_MAX_STATES;a++) {
                c = dwm_tokenize_line(c);
                /* Error */
                if(c == 0) {
                        return;
                }
                /* Success */
                if(*c == 0) {
                        return;
                }
        }
}

/* Fatal error while parsing Mararc file */
void dwm_fatal(char *why) {
        dw_alog_number("Fatal error in dwood2rc file on line ",dwm_linenum,
		why);
        exit(1);
}

/* Given an input character, and a character class (a number from one
 * to 32), return -1 if we don't match, 1 if we do match */
int dwm_char_class(int32_t in, int cclass) {
        /* In order to save space, we use this C bit of messyness:
         * condition ? true : false
         * This is normally against coding style; we do it so the function
         * fits within the 52-line-per-function limit */
        switch(cclass) {
                case 1: /* A-Za-z_ */
                        return dwm_is_alpha(in) ? 1 : -1;

                case 2: /* A-Za-z0-9_ */
                        return dwm_is_alphanum(in) ? 1 : -1;

                case 4: /* A-Za-z0-9_- */
                        return dwm_is_dname(in) ? 1 : -1;

                case 9: /* printable ASCII except # and " */
                        return dwm_is_instring(in) ? 1 : -1;

                case 14: /* 0-9 */
                        return dwm_is_number(in) ? 1 : -1;

                case 18: /* \r */
                        return (in == '\r') ? 1 : -1;

                case 19: /* A-Za-z0-9 */
                        return dwm_is_dnamestart(in) ? 1 : -1;

                case 20: /* \n */
                        return (in == '\n') ? 1 : -1;

                case 23: /* ' ' or \t */
                        return dwm_is_whitespace(in) ? 1 : -1;

                case 24: /* printable ASCII/hi-bit or \t */
                        return dwm_is_any(in) ? 1 : -1;

                case 25: /* A-Za-z */
                        return dwm_is_alphastart(in) ? 1 : -1;

                default:
                        return -1;
        }
}

/* Given an input character, and a state, return an action (upper 16 bits)
 * and a new state (lower 16 bits).
 * Input: An input character and a current state
 * Output: ((action << 16) | newstate) or -1 on error */
int32_t dwm_process_character(int32_t in, int32_t state) {
        uint16_t action = 0;
        uint16_t newstate = 0;
        int32_t match = -1;
        int a = 0, b = 0;

        /* Sanity checks */
        if(state >= DWM_MAX_STATES) {
                return -1;
        }

        if(fsm[state].pattern[0] == 0) { /* Invalid state */
                return -1;
        }

        for(a=0;a<DWM_MAX_PATTERNS && fsm[state].pattern[a] != 0;a++) {
                b = fsm[state].pattern[a];
                if(b == 0) {
                        return -1;
                /* patterns between 1 and 32 are multi-character classes,
                 * such as "letters" */
                } else if(b < 32) {
                        match = dwm_char_class(in,b);
                /* Higher numbered "patterns" are one-character patterns,
                 * with the value of the character being the ASCII character
                 * we match against */
                } else {
                        if(in == b) {
                                match = 1;
                        } else {
                                match = -1;
                        }
                }
                if(match == 1) {
                        action = fsm[state].action[a];
                        newstate = fsm[state].newstate[a];
                        break;
                }
        }

        if(fsm[state].pattern[a] == 0) { /* Invalid state */
                return -1;
        }

        if(newstate >= DWM_MAX_STATES || action > 4096) {
                return -1;
        }

        return (action << 16) | newstate;
}

/* Initialize the Mararc parameters */

void dwm_init_mararc() {
        int a = 0;
        for(a = 0; a < KEY_S_COUNT; a++) {
                key_s[a] = 0;
        }
        for(a = 0; a < KEY_D_COUNT; a++) {
                key_d[a] = 0;
        }
        /* Numeric mararc variables have default values.  Note that
         * changing these values does not always change the default
         * value; look for get_key_n in programs for some default 
         * values */
        key_n[DWM_N_maxprocs] = 32;
        key_n[DWM_N_timeout_seconds] = 3;
        key_n[DWM_N_dns_port] = 53;
        key_n[DWM_N_upstream_port] = 53;
        key_n[DWM_N_handle_overload] = 1;
        key_n[DWM_N_handle_noreply] = 1;
        key_n[DWM_N_recurse_min_bind_port] = 15000;
        key_n[DWM_N_recurse_number_ports] = 4096;
        key_n[DWM_N_hash_magic_number] = 1;
	key_n[DWM_N_maximum_cache_elements] = 1024;
	key_n[DWM_N_maradns_uid] = 99;
	key_n[DWM_N_maradns_gid] = 99;
	key_n[DWM_N_resurrections] = 1;
	key_n[DWM_N_num_retries] = 1;
	key_n[DWM_N_verbose_level] = 3;
	key_n[DWM_N_deliver_all] = 1;
	key_n[DWM_N_max_ttl] = 86400;
}

/* Look for a Mararc parameter; -1 if not found/error; 0-n if found
 * (0: First index, 1: second index, etc.) */
int32_t dwm_grep_params(dw_str *seek, char *list[], int max) {
        int e = 0;
        dw_str *look = 0;
        int ret = 0;

        if(seek == 0 || list == 0 || max < 1) {
                ret = -1;
                goto catch_dwm_grep_params;
        }

        for(e = 0; e < max; e++) {
                if(list[e] == 0) {
                        ret = -1;
                        goto catch_dwm_grep_params;
                }
                look = dw_create(64);
                dw_qrappend((uint8_t *)list[e],look,0);
                if(dw_issame(seek,look) == 1) {
                        ret = e;
                        goto catch_dwm_grep_params;
                }
                if(look != 0) {
                        dw_destroy(look);
                        look = 0;
                }
        }

        ret = -1;

catch_dwm_grep_params:
        if(look != 0) {
                dw_destroy(look);
                look = 0;
        }
        return ret;
}

/* Based on the actions done, determine what to do (this is called from
 * dwm_set_keys()).
 * Output:
 * -1: Error (bad action)
 * 0: Do nothing
 * 1: Set normal string variable
 * 2: Append to normal string variable
 * 3: Set dictionary variable
 * 4: Append to dictionary variable
 * 5: Set numeric variable
 */

int dwm_set_todo(dw_str **actions) {
        if(actions[1] == 0 || actions[6] != 0) { /* Do nothing */
                return 0;
        }
        if(actions[3] == 0) { /* No string to set */
                if(actions[4] == 0) { /* No number to set */
                        return -1;
                } else if(actions[2] != 0) { /* Dictionary variable */
                        return -1;
                } else if(actions[5] != 0) { /* += instead of = */
                        return -1; /* Not supported; should we support this? */
                } else {
                        return 5;
                }
        }
        if(actions[2] != 0) { /* Dictionary variable */
                if(actions[6] != 0) { /* foo["."] = {} line */
                        return 0;
                }
                if(actions[5] != 0) { /* += instead of = */
                        return 4;
                }
                return 3;
        }
        if(actions[5] != 0) { /* += instead of = */
                return 2;
        }
        return 1;
}

/* Based on the actions done, set the appropriate Mararc parameters */
void dwm_set_keys(dw_str **actions) {
        int todo = 0, num = 0;
        dw_str **list = 0;

        todo = dwm_set_todo(actions);
        if(todo == 0) {
                return;
        } else if(todo == -1) {
                dwm_fatal("Bad deadwoodrc action");
        }

        if(todo == 1 || todo == 2) { /* Normal variable */
                num = dwm_grep_params(actions[1],key_s_names,KEY_S_COUNT);
                list = key_s;
        } else if(todo == 3 || todo == 4) { /* Dictionary variable */
                num = dwm_grep_params(actions[1],key_d_names,KEY_D_COUNT);
                list = key_d;
        } else if(todo == 5) { /* Numeric variable */
                int32_t val;
                num = dwm_grep_params(actions[1],key_n_names,KEY_N_COUNT);
                val = dw_atoi(actions[4],0,10);
                if(val == -1) {
                        dwm_fatal("Invalid numeric value");
                }
                if(num < 0) { /* Make sure the parameter is a legal one */
                        dwm_fatal("Unknown deadwoodrc parameter");
                }
                key_n[num] = val;
                return;
        } else { /* Shouldn't get here */
                return;
        }

        if(list == 0) { /* Sanity check */
                dwm_fatal("Unknown deadwoodrc action");
        }
        if(num < 0) { /* Make sure the parameter is a legal one */
                dwm_fatal("Unknown deadwoodrc parameter");
        }

        if(list[num] == 0 && (todo & 1) == 0) {
                dwm_fatal("Appending to unset parameter");
        } else if(list[num] != 0 && (todo & 1) == 0 /* append */) {
                dw_append(actions[3],list[num]);
                return;
        } else if(list[num] != 0) {
                dwm_fatal("Deadwoodrc parameter set twice");
        }

        list[num] = dw_copy(actions[3]);
}

/* If there is an action to perform, perform the action */
void dwm_do_action(int32_t ch, int32_t action, dw_str **actions) {
        if(actions[action] == 0) {
                /* This function is *only* called from dwm_parse_line,
                 * and the destructor for these strings is at the
                 * end of that function */
                actions[action] = dw_create(384);
        }
        dw_addchar(ch, actions[action]);
}

/* Parse a line in a MaraRC file */
int dwm_parse_line(FILE *look) {
        dw_str *actions[11];
        int a = 0;
        int32_t action = 0, state = 0, mix = 0, ch = 0, last = 0;
        int ret = 0;

        for(a = 0; a < 11; a++) {
                actions[a] = 0;
        }

        ch = fgetc(look);
        if(feof(look)) {
                ret = 1; /* End of file reached */
                goto catch_dwm_parse_line;
        }

        while(action != 10) {
                mix = dwm_process_character(ch,state);
                if(mix < 0) {
                        ret = -1;
                        goto catch_dwm_parse_line;
                }
                action = (mix >> 16) & 0x7fff;
                state = mix & 0x7fff;
                if(action == 10) { /* Terminate */
                        ret = 0; /* Line parsed */
                        break; /* Now we need to set the keys */
                } else if(action > 0 && action < 10) {
                        dwm_do_action(ch,action,actions);
                }
                last = ch;
                ch = fgetc(look);
                if(feof(look) && last != '\r' && last != '\n') {
                        ret = -2; /* Premature EOF */
                        goto catch_dwm_parse_line;
                }
        }

        dwm_set_keys(actions); /* Sets mararc parameters based on actions */

catch_dwm_parse_line: /* Actions strings destructor */
        for(a = 0; a < 11; a++) {
                if(actions[a] != 0) {
                        dw_destroy(actions[a]);
                        actions[a] = 0;
                }
        }
        return ret;
}

/* Parse a mararc file; this should only be called once when executing
 * deadwood.  Note that this is the *only* public method in this entire
 * file; all other functions in this file should only be called from
 * other functions in this file.
 * Input: c string that points to mararc file
 * Output: 1 on success; program exits on mararc parse error */
int dwm_parse_mararc(char *name) {
        FILE *look;
        int a = 0;

        dwm_linenum = 1;

        dwm_init_fsm();
        dwm_init_mararc();

        look = fopen(name,"rb");
        if(look == NULL) {
                return -1; /* File open error */
        }

        do {
                a = dwm_parse_line(look);
                if(a != 0 && a != 1) {
                        dwm_fatal("deadwoodrc parse error");
                }
                dwm_linenum++;
        } while (a == 0);

        fclose(look);

        return 1;
}

#ifdef HAVE_MAIN

/* Debugging function to make sure fsm is correctly set up */

void show_fsm() {
        int a = 0, b = 0, q = 0;
        for(a = 0; a < DWM_MAX_STATES; a++) {
                if(fsm[a].pattern[0] != 0) {
                        printf("State %d\n",a);
                }
                for(b = 0; b < DWM_MAX_PATTERNS; b++) {
                        q = fsm[a].pattern[b];
                        if(q > 0 && q < 32 && fsm[a].action[b] != 10) {
                                printf("Pattern @%d action %d newstate %d\n",
                                        fsm[a].pattern[b],
                                        fsm[a].action[b],
                                        fsm[a].newstate[b]);
                        } else if(q >= '!' && fsm[a].action[b] != 10) {
                                printf("Pattern %c action %d newstate %d\n",
                                        fsm[a].pattern[b],
                                        fsm[a].action[b],
                                        fsm[a].newstate[b]);

                        } else if(q > 0 && q < 32) {
                                printf("Pattern @%d action %d (terminate)\n",
                                        fsm[a].pattern[b],
                                        fsm[a].action[b]);
                        } else if(q >= '!') {
                                printf("Pattern %c action %d (terminate)\n",
                                        fsm[a].pattern[b],
                                        fsm[a].action[b]);
                        } else {
                                break;
                        }
                }
                if(fsm[a].pattern[0] != 0) {
                        printf("\n");
                }
        }
}

int main() {
        int a = 0;

        dwm_parse_mararc("deadwoodrc");

        show_fsm();

        for(a=0;a<KEY_S_COUNT;a++) {
                printf("%s is ",key_s_names[a]);
                if(key_s[a] == 0) {
                        printf("not set.\n");
                } else {
                        dw_stdout(key_s[a]);
                }
        }

        for(a=0;a<KEY_D_COUNT;a++) {
                printf("%s is ",key_d_names[a]);
                if(key_s[a] == 0) {
                        printf(" not set.\n");
                } else {
                        dw_stdout(key_d[a]);
                }
        }

        return 0;
}

#endif /* HAVE_MAIN */

