/* Copyright (c) 2007 Sam Trenholme
 *
 * TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

#ifndef __DW_SOCKET_DEFINED__
#define __DW_SOCKET_DEFINED__

#include "DwStr.h"
#include "DwMararc.h"
#include "DwSys.h"
#include "DwStr_functions.h"
#include "DwRadioGatun.h"
#include "DwHash.h"

#ifdef __CYGWIN__
#define MSG_DONTWAIT 0
#endif /* __CYGWIN__ */
#ifndef MINGW
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <grp.h>
#else  /* MINGW */
/* Maximum number of open sockets; make this big like it is in *NIX */
#define FD_SETSIZE 512
#include <winsock.h>
#include <wininet.h>
#endif /* MINGW */
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>

#ifdef MINGW
#define NO_INET_PTON
#define sa_family_t uint16_t
#define socklen_t int32_t
#define EINPROGRESS WSAEWOULDBLOCK
/* This has a strange name because winsock.h defines SOCKET_ERROR */
#define SCKT_ERR WSAGetLastError()
/* This actually isn't supported in winsock but we do the ioctl thing
 * to make the socket non-blocking, so this should be a non-issue */
#define MSG_DONTWAIT 0
#define make_socket_nonblock(a) dont_block=1;ioctlsocket(a,FIONBIO,&dont_block)
#else /* MINGW */ /* The following are defined for *NIX */
#define SCKT_ERR errno
#define SOCKET int
#define INVALID_SOCKET -1
#define closesocket(a) close(a)
#define make_socket_nonblock(a) fcntl(a,F_SETFL,O_NONBLOCK)
#endif /* MINGW */

/* Parameter that is currently hardcoded in the source code */
#define DW_MAXIPS 7    /* Maximum number of bind addresses or upstream IPs */

/* Structures for storing IP addresses */
typedef struct {
        uint8_t len;
#ifdef IPV6
        uint8_t ip[17];
#else /* IPV6 */
        uint8_t ip[5];
#endif /* IPV6 */
} ip_addr_T;

/* Structure for storing an IP address *and* a netmask; note that the
 * beginning of the structure *must* have the exact same form as the
 * ip_addr_T structure above */
typedef struct {
        uint8_t len;
#ifdef IPV6
        uint8_t ip[17];
        uint8_t mask[17];
#else /* IPV6 */
        uint8_t ip[5];
        uint8_t mask[5];
#endif /* IPV6 */
} ip_mask_T;

/* storage for both sockaddr_in and sockaddr_in6; note that this needs
 * to be in that format that system calls like bind() and what not use
 * for the sockaddr format (16-bit family, followed by the IP) */
typedef struct {
        union {
                sa_family_t family;
                struct sockaddr_in v4;
#ifdef IPV6
                struct sockaddr_in6 v6;
#endif
        } u;
} sockaddr_all_T;

/* shortcuts for above struct */
#define V4 u.v4
#define V6 u.v6
#define Family u.v4.sin_family

/* A pending remote connection */
typedef struct {
        SOCKET socket;
        SOCKET from_socket;
        int64_t die;
        ip_addr_T ip;
        uint16_t port;
        uint16_t local_id; /* Query ID generated by DNS server connecting
                            * to Deadwood */
        uint16_t remote_id; /* Query ID generated by Deadwood */
	int retries; /* Number of times to retry connection */
	dw_str *query;
} remote_T;

/* Make this also look like a function; this resets a given remote
 * connection */

#define reset_rem(a) rem[a].socket = INVALID_SOCKET; \
                     rem[a].from_socket = INVALID_SOCKET; \
                     rem[a].die = 0; \
                     rem[a].ip.len = 0; \
                     rem[a].port = 0; \
                     rem[a].local_id = 0; \
                     rem[a].remote_id = 0; \
		     rem[a].retries = num_retries; \
		     if(rem[a].query != 0) { \
			dw_destroy(rem[a].query); \
		        rem[a].query = 0; } 

/* Function parameters */

/* The following functions are private, and should only be called from
 * DwSocketUdp.c and DwSocketTcp.c */

/* In DwSocket.c */

/* Given a "client" that recvfrom/accept gave us (which has the IP and port
 * number hidden in it), extract the IP and port from that "client", put the
 * IP information in from_ip, and return a 16-bit number that has the port
 * the query came from */
uint16_t get_from_ip_port(ip_addr_T *from_ip, sockaddr_all_T *client);

/* See if a given ip is a permitted IP in the recursive_acl list.  Input:
 * IP we are checking.  Output: -1 on error; 0 if not permitted; 1 if
 * permitted */
int check_ip_acl(ip_addr_T *ip);

/* Configure the dns_do local bind structure.  Given an IP we
 * want to bind to, fill up the dns_udp structure with that IP
 * and set things up to bind to the dns_port port (normally 53, DNS).
 * Establish a socket for binding to, and return the value of the
 * socket.  type is SOCK_DGRAM for a UDP socket; SOCK_STREAM for TCP */
SOCKET bind_set_dns(ip_addr_T *ip, sockaddr_all_T *dns_do, int type);

/* Given an IP address to bind to, and a type of port it is (SOCK_DGRAM for
 * a UDP port; SOCK_STREAM for a TCP port), bind to that port and return the
 * socket number corresponding to that port */
SOCKET do_bind(ip_addr_T *ip,int type);

/* Get an upstream IP address (Not thread safe w/o locking) */
ip_addr_T get_upstream_ip();

/* Find a free remote pending connection */
int32_t find_free_remote();

/* In DwUdpSocket.c */
/* Get and process a local DNS request */
void get_local_udp_packet(SOCKET sock);

/* Get and process a remote DNS request */
void get_remote_udp_packet(int b, SOCKET sock);

/* Here are all of the public functions in this program; anything else
 * is private and should not be called outside of DwSocket.c */

/* Read mararc parameters and set global variables based on those
 * parameters */
void process_mararc_params();

/* Bind to all IP addresses we are to bind to and return the number of
 * IP addresses we got */
int bind_all_udp();

/* Initialize the list of pending remote replies */
void init_b_remote();

/* Main loop: Recieve from bound sockets, forward those on to upstream, and
 *            forward replies from upstream to bound sockets */
int bigloop();

/* The following functions are private functions, and should only be called
 * from DwSocket.c, DwUdpSocket.c, and DwTcpSocket.c */

/* Create a DNS header suitable for giving back to the client */
dw_str *make_dns_header(int32_t id, int16_t flags, int32_t ancount,
                        int32_t nscount, int32_t arcount);

/* Given two dw_strings with the question and answer, make a DNS packet
 * we can give back to the client; this will multilate answer so be
 * careful */
dw_str *make_dns_packet(dw_str *question, dw_str *answer, int32_t id);

/* Given a remote number, a C-string with the packet to send them, and
 * the length of that string, make a connection to a remote server */
void make_remote_connection(int32_t n, unsigned char *packet, int len);

/* Forward a remote reply back to the client */
void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip);

/* Send a server failure back to the client when there is no reply from
 * the upstream server.  Input: The pending remote connection number. */
void server_fail_noreply(int a);

#endif /* __DW_SOCKET_DEFINED__ */
