/* Copyright (c) 2007-2009 Sam Trenholme, Jean-Jacques Sarton, and 
 * Marko Njezic
 *
 * TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

#include "DwSocket.h"
#include <signal.h>
#include <errno.h>

/* One parameter that may eventually become a dwood2rc parameter */
#define TCP_BUFFERSIZE 1024

/* Mararc parameters that are set in DwMararc.c */
extern dw_str *key_s[];
extern dw_str *key_d[];
extern int32_t key_n[];

/* Parameters set in DwSys.c */
extern int64_t the_time;
extern dwr_rg *rng_seed;

/* List of addresses we will bind to */
extern ip_addr_T bind_address[];
extern ip_addr_T upstream_address[];

/* Some global variables */
extern int maxprocs;
extern int timeout_seconds;
extern int dns_port;
extern int upstream_port;
extern int num_retries;

/* Some tcp-specific structures */
typedef struct {
        SOCKET local;
        SOCKET remote;
	char *buffer;
	SOCKET buffer_socket;
	ssize_t buffer_len;
        int64_t die;
        } tcp_remote_T;

#define init_tcp_rem(a) tcp_rem[a].local = INVALID_SOCKET; \
                         tcp_rem[a].remote = INVALID_SOCKET; \
			 tcp_rem[a].buffer = 0; \
			 tcp_rem[a].buffer_socket = INVALID_SOCKET; \
			 tcp_rem[a].buffer_len = 0; \
                         tcp_rem[a].die = 0;

#define reset_tcp_rem(a) tcp_rem[a].local = INVALID_SOCKET; \
                         tcp_rem[a].remote = INVALID_SOCKET; \
			 if(tcp_rem[a].buffer != 0) { \
				free(tcp_rem[a].buffer); \
			 	tcp_rem[a].buffer = 0; } \
			 tcp_rem[a].buffer_socket = INVALID_SOCKET; \
			 tcp_rem[a].buffer_len = 0; \
                         tcp_rem[a].die = 0;

tcp_remote_T *tcp_rem;
int tcp_b_local[DW_MAXIPS + 1]; /* Local sockets */

/* The following is needed because Winsock's "make this socket non-blocking"
 * uses a pointer to a number as one of its arguments */
#ifdef MINGW
extern u_long dont_block;
extern void windows_socket_start();
#endif /* MINGW */

/* The upstream server we will connect to (round robin rotated) */

/* Allocate the memory for the list of the open remote TCP connections.
 * This memory is never freed once allocated, because this data is always
 * used by the program */
void malloc_tcp_remote() {
        tcp_rem = malloc((maxprocs + 1) * sizeof(tcp_remote_T));
        if(tcp_rem == 0) {
                dw_alog_3strings("Fatal: Could not allocate tcp_rem","","");
                exit(1);
        }
}

/* Initialize the values of all the open remote TCP connections */
void init_tcp_b_remote() {
        int a = 0;
        for(a = 0; a < maxprocs; a++) {
                init_tcp_rem(a);
        }
}

/* TCP bind to all IP addresses we are to bind to and return the number of
 * IP addresses we got */
int bind_all_tcp() {
        int a = 0;
        int count = 0;
        for(a = 0; a < DW_MAXIPS; a++) {
                if(bind_address[a].len != 0) {
                        tcp_b_local[a] = do_bind(&bind_address[a],SOCK_STREAM);
                        if(tcp_b_local[a] != -1) {
                                count++;
                        } else {
                                return -1;
                        }
                } else {
                        tcp_b_local[a] = -1;
                }
        }
        return count;
}

/* Get the maximum TCP socket number or either a local or remote socket;
 * add one to the result so it can be used by select() */
int tcp_get_max() {
        int a = 0;
        int max = 0;
        for(a = 0; a < DW_MAXIPS; a++) {
                if(tcp_b_local[a] > max) {
                        max = tcp_b_local[a];
                }
        }
        for(a = 0; a < maxprocs; a++) {
                if(tcp_rem[a].local != INVALID_SOCKET) {
                        if(tcp_rem[a].local > max) {
                                max = tcp_rem[a].local;
                        }
                        if(tcp_rem[a].remote > max) {
                                max = tcp_rem[a].remote;
                        }
                }
        }
        /* Socket lib quirk: Select uses max + 1, not max */
        return max + 1;
}

/* Set the rx_fd list, which is used by select() */
void tcp_set_rx_fd(fd_set *rx_fd) {
        int a = 0;
        FD_ZERO(rx_fd);
        for(a = 0; a < DW_MAXIPS; a++) {
                if(tcp_b_local[a] != -1) {
                        FD_SET(tcp_b_local[a],rx_fd);
                }
        }
        for(a = 0; a < maxprocs; a++) {
                if(tcp_rem[a].local != INVALID_SOCKET) {
                        FD_SET(tcp_rem[a].local,rx_fd);
                        if(tcp_rem[a].remote != INVALID_SOCKET) {
                                FD_SET(tcp_rem[a].remote,rx_fd);
                        }
                }
        }
}

/* Find a free remote/local TCP connection to use; return -1 if
 * there isn't one (we're overloaded) */
int32_t find_free_tcp_remote() {
        int32_t a = 0;
        for(a = 0; a < maxprocs; a++) {
                if(tcp_rem[a].local == INVALID_SOCKET) {
                        return a;
                }
        }
        return -1; /* None available (we're overloaded) */
}

/* Set up a TCP server that we will use to connect to a remote host */
SOCKET setup_tcp_server(sockaddr_all_T *server) {
        SOCKET remote = INVALID_SOCKET;
        ip_addr_T rem_ip = {0,{0,0}};
        rem_ip = get_upstream_ip();
        if(rem_ip.len == 4) {
                server->V4.sin_family = AF_INET;
                server->V4.sin_port = htons(upstream_port);
                memcpy(&(server->V4.sin_addr),rem_ip.ip,4);
                remote = socket(AF_INET,SOCK_STREAM,0);
#ifdef IPV6
        } else if(rem_ip.len == 16) {
                server->V6.sin6_family = AF_INET6;
                server->V6.sin6_port = htons(upstream_port);
                memcpy(&(server->V6.sin6_addr),rem_ip.ip,16);
                remote = socket(AF_INET6,SOCK_STREAM,0);
#endif /* IPV6 */
        } else {
                return INVALID_SOCKET;
        }
        return remote;
}

/* Given a tcp socket s, accept the connection on that socket, then
 * open up a TCP connection to the upstream server, and assign a
 * tcp_rem with the socket numbers for the accepted local connection
 * and the connection to the upstream server */
void local_tcp_accept(int s) {
        sockaddr_all_T client, server;
        int local = 0;
        int remote = 0;
        socklen_t len = sizeof(struct sockaddr_in);
        int b = 0;
        ip_addr_T from_ip;

        b = find_free_tcp_remote();
        if(b == -1) { /* Out of active TCP connections */
                return;
        }

        len = sizeof(client);
        local = accept(s,(struct sockaddr *)&client,&len);
        make_socket_nonblock(local);

        if(local == INVALID_SOCKET) { /* accept() error */
                return;
        }

        /* This is where we do ip-based packet rejection */
        get_from_ip_port(&from_ip,&client);
        if(check_ip_acl(&from_ip) != 1) {
                closesocket(local);
                return;
        }

        remote = setup_tcp_server(&server);
        if(remote == INVALID_SOCKET) {
                closesocket(local);
                return;
        }

        make_socket_nonblock(remote);
#ifdef IPV6
        if (server.Family == AF_INET6)
                len = sizeof(struct sockaddr_in6);
#endif
        if(connect(remote,(struct sockaddr *)&server,len) == -1) {
                if(SCKT_ERR != EINPROGRESS) { /* connect() takes time */
                        closesocket(local);
                        closesocket(remote);
                        return;
                } 
        }
        tcp_rem[b].local = local;
        tcp_rem[b].remote = remote;
        tcp_rem[b].die = get_time() + ((int64_t)timeout_seconds << 8);
}

/* Attempt to flush buffered data for a given pending TCP connection */
int tcp_flush_buffer(int r) {
	ssize_t actual = 0;
	if(tcp_rem[r].buffer_socket == INVALID_SOCKET) {
		return 0;
	}
	if(tcp_rem[r].buffer == 0) {
		return -1; /* Error: If there's a socket to send the buffer,
                              there better be buffered data to send */
        }
	
	actual = send(tcp_rem[r].buffer_socket,
                      tcp_rem[r].buffer,
		      tcp_rem[r].buffer_len,MSG_DONTWAIT);
	
	if(actual != tcp_rem[r].buffer_len) {
		if(actual != -1) {
			dw_log_string(
				"WARNING: Partial buf sends not supported",0);
		}
		return -1;
	}

	tcp_rem[r].buffer_socket = INVALID_SOCKET;
	free(tcp_rem[r].buffer);
	tcp_rem[r].buffer = 0;
	tcp_rem[r].buffer_len = 0;
	return 0;
	
}

/* When the local code has sent bytes over a TCP pipe, pipe those bytes
 * to the remote TCP connection */
void tcp_local2remote(int local,int remote,int r) {
        char buffer[TCP_BUFFERSIZE + 1];
        ssize_t len = 0;
	ssize_t actual = 0;

	/* Don't accept data until buffers are cleared */
	if(tcp_rem[r].buffer_socket != INVALID_SOCKET &&
		tcp_flush_buffer(r) != 0) {
		return;
	}

        len = recv(local,buffer,TCP_BUFFERSIZE,MSG_DONTWAIT);

        if(len <= 0) {
                closesocket(tcp_rem[r].local);
                closesocket(tcp_rem[r].remote);
                reset_tcp_rem(r);
                return;
        }

        /* Note: Winsock might not send all of the data; *NIX may have
         * the same problem */
	actual = send(remote,buffer,len,MSG_DONTWAIT);
        if(actual != len) { 
		void *nowarn;
		if(actual != -1) { /* Partial sends not supported */
			dw_log_string("WARNING Partial sends not supported",0);
                	closesocket(tcp_rem[r].local);
                	closesocket(tcp_rem[r].remote);
                	reset_tcp_rem(r);
                	return;
		}
		/* Buffer the data */
		tcp_rem[r].buffer_socket = remote;
		tcp_rem[r].buffer = (char *)malloc(len + 2);
		nowarn = memcpy(tcp_rem[r].buffer, buffer, len);
		tcp_rem[r].buffer_len = len;
        }
        /* We only kill an inactive connection */
        tcp_rem[r].die = get_time() + ((int64_t)timeout_seconds << 8);
}

#define tcp_remote2local(a,b,c) tcp_local2remote(a,b,c)

/* Disconnect idle TCP connections */
void kill_tcp_expired() {
        int a = 0;
        for(a = 0; a < maxprocs; a++) {
                if(tcp_rem[a].die > 0 && tcp_rem[a].die < get_time()) {
                        closesocket(tcp_rem[a].local);
                        closesocket(tcp_rem[a].remote);
                        reset_tcp_rem(a);
                }
        }
}

/* Process any pending connections which select() caught */
void tcp_process_results(int a, fd_set *rx_fd) {
        int b = 0, z = 0;

        /* Find the pending connection */
        while(a > 0 && z < 10000) {
                /* Handle new connections */
                for(b = 0; b < DW_MAXIPS; b++) {
                        if(tcp_b_local[b] != INVALID_SOCKET &&
                           FD_ISSET(tcp_b_local[b],rx_fd)) {
                                local_tcp_accept(tcp_b_local[b]);
                                a--;
                                if(a <= 0) {
                                        return;
                                }
                        }
                }
                /* Handle already-established connections */
                for(b = 0; b < maxprocs; b++) {
			if(tcp_rem[b].buffer_socket != INVALID_SOCKET) {
				 tcp_flush_buffer(b);
			}
                        if(tcp_rem[b].local != INVALID_SOCKET &&
                           FD_ISSET(tcp_rem[b].local,rx_fd)) {
                                tcp_local2remote(tcp_rem[b].local,
                                                 tcp_rem[b].remote,b);
                                a--;
                                if(a <= 0) {
                                        return;
                                }
                        }
                        if(tcp_rem[b].remote != INVALID_SOCKET &&
                           FD_ISSET(tcp_rem[b].remote,rx_fd)) {
                                tcp_remote2local(tcp_rem[b].remote,
                                                 tcp_rem[b].local,b);
                                a--;
                                if(a <= 0) {
                                        return;
                                }
                        }
                }
                z++;
        }
}

#ifdef MINGW
extern int run_loop;
#endif /* MINGW */

void tcp_bigloop() {
        int a = 0, b = 0;
        int max = 0;
        fd_set rx_fd;
        struct timeval timeout;

#ifndef MINGW
        for(;;) {
#else /* MINGW */
	while (run_loop == 1) {
#endif /* MINGW */
                set_time();
                max = tcp_get_max();
                tcp_set_rx_fd(&rx_fd);
                timeout.tv_sec = 0;
                timeout.tv_usec = 50000; /* 20 times a second */
                a = select(max,&rx_fd,NULL,NULL,&timeout);

                /* If we have results, process them */
                if(a > 0) {
                        tcp_process_results(a,&rx_fd);
                }
		/* Flush buffers */
                for(b = 0; b < maxprocs; b++) {
			if(tcp_rem[b].buffer_socket != INVALID_SOCKET) {
				 tcp_flush_buffer(b);
			}
		}
                /* Kill off expired pending connection */
                kill_tcp_expired();
        }
}

/* This stuff should be removed; duplication of code elsewhere */
#ifdef MINGW
void dw_tcp_windows_socket_start() {
        WSADATA wsaData;
        WORD wVersionRequested = MAKEWORD(2,2);
        WSAStartup( wVersionRequested, &wsaData);
}
#endif /* MINGW */

/* Show them usage then exit; used when they don't put in correct
 * command line parameters */
void dw_tcp_usage() {
        printf("Usage: DwTcp [-f dwood2rc]\n");
        exit(1);
}


/* Main(): run the program */
int dw_tcp_main(int argc, char **argv) {
        int bind_count = 0; /* Number of IPs we bound to */
#ifdef IPV6
        dw_alog_3strings("DwTcp: A DNS TCP load balancer (IPv6 supported)",
		"","");
#else /* IPV6 */
        dw_alog_3strings("DwTcp: A DNS TCP load balancer (IPv4-only)","","");
#endif /* IPV6 */
#ifndef MINGW
        /* *NIX bug: TCP connections, in certain circumstances, can cause
         * the program to get a SIGPIPE signal, which, by default, terminates
         * the program.  This is a security gotcha (DJBdns' dnscache, for
         * example, has a denial-of-service security problem since it doesn't
         * catch this signal) */
        signal(SIGPIPE, SIG_IGN);
#endif /* MINGW */
        /* Process the command-line parameters and dwood2rc file */
        if(argc != 1 && argc != 3) {
                dw_tcp_usage();
        } else if(argc == 3) {
                if(argv[1] == 0 || argv[2] == 0) {
                        dw_tcp_usage();
                } if(*(argv[1]) != '-' || *(argv[1] + 1) != 'f' ||
                     *(argv[1] + 2) != 0) {
                        dw_tcp_usage();
                }
                process_mararc(argv[2]);
        } else {
#ifndef MINGW
                process_mararc("/etc/dwood2rc");
#else /* MINGW */
		process_mararc("dwood2rc.txt");
#endif /* MINGW */
        }
        process_mararc_params();
#ifdef MINGW
        windows_socket_start();
#endif /* MINGW */
        bind_count = bind_all_tcp();
        if(bind_count > 0) {
                dw_log_number("We bound to ",bind_count," addresses",1);
        } else {
                dw_fatal("Unable to bind to any IP addresses");
        }
#ifdef MINGW
        fflush(stdout);
#endif /* MINGW */
        init_rng();
        sandbox(); /* Enter a chroot() sandbox and drop root */
        /* Finish up initializing */
        malloc_tcp_remote();
        init_tcp_b_remote();
        /* And run the server */
        tcp_bigloop();
        return 0;
}

