/* Copyright (c) 2007-2012 Sam Trenholme 
 * IPv6 code contributed by Jean-Jacques Sarton
 *
 * TERMS
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This software is provided 'as is' with no guarantees of correctness or
 * fitness for purpose.
 */

#include "DwSocket.h"

/* Mararc parameters that are set in DwMararc.c */
extern dw_str *key_s[];
extern dw_str *key_d[];
extern int32_t key_n[];

/* Parameters set in DwSys.c */
extern int64_t the_time;
extern dwr_rg *rng_seed;
extern dw_hash *cache;

/* List of addresses we will bind to */
extern ip_addr_T bind_address[];
extern ip_addr_T upstream_address[];

/* List of active sockets */
extern SOCKET b_local[];
extern SOCKET *b_remote;

/* The list of pending remote connections */
extern remote_T *rem;

/* The numeric mararc parameters */
extern int maxprocs;
extern int timeout_seconds;
extern int dns_port;
extern int upstream_port;
extern int handle_overload;
extern int resurrections;
extern int min_bind;
extern int num_ports;
extern int num_retries;
extern int deliver_all;
extern int32_t max_ttl;

#ifdef MINGW
/* Needed for the Windows way of making a socket non-blocking */
extern u_long dont_block;
#endif /*MINGW*/

/* The upstream server we will connect to (round robin rotated) */

/* Bind to all IP addresses we are to bind to and return the number of
 * IP addresses we got or INVALID_SOCKET on error */
int bind_all_udp() {
        int a = 0;
        int count = 0;
        for(a = 0; a < DW_MAXIPS; a++) {
                if(bind_address[a].len != 0) {
                        b_local[a] = do_bind(&bind_address[a],SOCK_DGRAM);
                        if(b_local[a] != INVALID_SOCKET) {
                                count++;
                        } 
                } else {
                        b_local[a] = INVALID_SOCKET;
                }
        }
        return count;
}

/* Create a sockaddr_in that will be bound to a given port; this is
 * used by the code that binds to a randomly chosen port */
void setup_bind(sockaddr_all_T *dns_udp, uint16_t port, int len) {
        if(dns_udp == 0) {
                return;
        }
        memset(dns_udp,0,sizeof(dns_udp));
        if(len == 4) {
                dns_udp->V4.sin_family = AF_INET;
                dns_udp->V4.sin_addr.s_addr = htonl(INADDR_ANY);
                dns_udp->V4.sin_port = htons(port);
#ifdef IPV6
        } else if(len == 16) {
                dns_udp->V6.sin6_family = AF_INET6;
                dns_udp->V6.sin6_addr = in6addr_any;
                dns_udp->V6.sin6_port = htons(port);
#endif
        } else { /* Bad ip */
                return;
        }
        return;
}

/* This tries to bind to a random port up to 10 times; should it fail
 * after 10 times, it returns a -1 */
int do_random_bind(SOCKET s, int len) {
        sockaddr_all_T dns_udp;
        int a = 0;
        int success = 0;
        socklen_t len_inet = sizeof(struct sockaddr_in);

        for(a = 0; a < 10; a++) {
                /* Set up random source port to bind to */
                setup_bind(&dns_udp,
                           min_bind + (dwr_rng(rng_seed) & num_ports), len);
#ifdef IPV6
                if (dns_udp.Family == AF_INET6)
                        len_inet = sizeof(struct sockaddr_in6);
#endif
                /* Try to bind to that port */
                if(bind(s, (struct sockaddr *)&dns_udp,
                   len_inet) != -1) {
                        success = 1;
                        break;
                }
        }
        if(success == 0) { /* Bind to random port failed */
                return -1;
        }
        return 1;
}

/* Create a sockaddr_in that will connect to a given address; -1 on fail;
 * this is used by the code that "connects" to the remote DNS server */
SOCKET setup_server(sockaddr_all_T *server, ip_addr_T *addr) {
        int s = INVALID_SOCKET;
        if(server == 0 || addr == 0) {
                return INVALID_SOCKET;
        }

        memset(server,0,sizeof(server));
        if ( addr->len == 4 ) {
                server->V4.sin_family = AF_INET;
                server->V4.sin_port = htons(upstream_port);
                memcpy(&(server->V4.sin_addr),addr->ip,4);
                s = socket(AF_INET,SOCK_DGRAM,0);
#ifdef IPV6
        } else if ( addr->len == 16 ) {
                server->V6.sin6_family = AF_INET6;
                server->V6.sin6_port = htons(upstream_port);
                memcpy(&(server->V6.sin6_addr),addr->ip,16);
                s = socket(AF_INET6,SOCK_DGRAM,0);
#endif
        } else {
                return INVALID_SOCKET;
        }
        return s;
}

/* Given a remote number, a C-string with the packet to send them, and
 * the length of that string, make a connection to a remote server */
void make_remote_connection(int32_t n, unsigned char *packet, int len) {
        sockaddr_all_T server;
        SOCKET s = 0;
        int32_t rnum = -1;
        ip_addr_T addr = {0,{0,0}};
        socklen_t inet_len = sizeof(struct sockaddr_in);

        if(rem[n].socket != INVALID_SOCKET) { /* Already used (sanity check) */
                return;
        }

        /* Get a random query ID to send to the remote server */
        rnum = set_dns_qid(packet,len,dwr_rng(rng_seed));
        if(rnum == -1) {
                return;
        }

        /* Get IP of remote server and open a socket to them */
        addr = get_upstream_ip();
        s = setup_server(&server,&addr);
        if(s == INVALID_SOCKET) { /* Failed to make socket */
                return;
        }

#ifdef IPV6
        if (server.Family == AF_INET6)
                 inet_len = sizeof(struct sockaddr_in6);
#endif

        make_socket_nonblock(s); /* Linux kernel bug */

        /* Bind to source port; "connect" to remote server; send packet */
        if ((do_random_bind(s,addr.len) == -1) ||
            (connect(s, (struct sockaddr *)&server, inet_len) == -1) ||
            (send(s,packet,len,0) < 0)) {
                closesocket(s);
                return;
        }

        /* OK, now note the pending connection */
        b_remote[n] = s;
        rem[n].socket = s;
        rem[n].remote_id = rnum;
}

/* Send a server failure back to the client when the server is overloaded.
 * This is a bit of a hack; we just take the client's DNS packet, and send 
 * it as-is with QR set to 1 (a response) and RCODE set to 2 (server 
 * failure) */
void send_server_fail(sockaddr_all_T *client,unsigned char *a, int len,
                      SOCKET sock) {
        socklen_t c_len = sizeof(struct sockaddr_in);

        if(len < 12) {
                return;
        }

#ifdef IPV6
        if (client->Family == AF_INET6)
                c_len = sizeof(struct sockaddr_in6);
#endif
        a[2] |= 0x80; /* Set QR (reply, not question) */
        a[3] &= 0xf0; a[3] |= 0x02; /* Set RCODE */

        sendto(sock,(void *)a,len,0,(struct sockaddr *)client, c_len);

}

/* Send a local DNS request to the upstream (remote) server; this
 * requires a few parameters having to do with the connection
 * to be passed to the function */
int forward_local_udp_packet(SOCKET sock, int32_t local_id,
     ip_addr_T *from_ip, uint16_t from_port, unsigned char *a,
     int len) {
        int32_t b = 0;

        b = find_free_remote();
        if(b == -1) { /* We're out of remote pending connections */
                return -1;
        }

        /* Make the remote connection */
        if(rem[b].socket != INVALID_SOCKET) { /* Sanity check */
                return -1;
        }
        make_remote_connection(b,a,len);
        if(rem[b].socket == INVALID_SOCKET) { /* Sanity check */
                return -1;
        }
        rem[b].from_socket = sock;
        rem[b].ip = *from_ip;
        rem[b].port = from_port;
        rem[b].local_id = local_id;
        rem[b].die = get_time() + ((int64_t)timeout_seconds << 8);
	rem[b].query = dw_get_dname_type(a,12,len);
        return 0;
}

/* Create a DNS header suitable for giving back to the client */
dw_str *make_dns_header(int32_t id, int16_t flags, int32_t ancount,
                        int32_t nscount, int32_t arcount) {
        dw_str *out = 0;

        if(id < 0 || ancount < 0 || nscount < 0 || arcount < 0) {
                return 0; /* Sanity check */
        }

        /* Destroyed after giving the client the reply */
        out = dw_create(515);

        /* Query ID; echo from client */
        if(dw_put_u16(out, id, -1) == -1) {
                goto catch_make_dns_header;
        }

        /* QR; Opcode; AA; TC; RD; RA; Z; RCODE */
        if(dw_put_u16(out, flags, -1) == -1) {
                goto catch_make_dns_header;
        }

        /* QDCOUNT = 1 */
        if(dw_put_u16(out,1,-1) == -1) {
                goto catch_make_dns_header;
        }

        /* And, finally, ANCOUNT, NSCOUNT, and ARCOUNT */
        if(dw_put_u16(out,ancount,-1) == -1 ||
           dw_put_u16(out,nscount,-1) == -1 ||
           dw_put_u16(out,arcount,-1) == -1) {
                goto catch_make_dns_header;
        }

        return out;

catch_make_dns_header:
        if(out != 0) {
                dw_destroy(out);
                out = 0;
        }
        return 0;
}

/* Given two dw_strings with the question and answer, make a DNS packet
 * we can give back to the client; this will multilate answer so be
 * careful */
dw_str *make_dns_packet(dw_str *question, dw_str *answer, int32_t id) {
        int32_t ancount = 0, nscount = 0, arcount = 0;
        int is_nxdomain = 0;
        dw_str *out = 0;

        is_nxdomain = dw_pop_u8(answer);
        arcount = dw_pop_u16(answer);
        nscount = dw_pop_u16(answer);
        ancount = dw_pop_u16(answer);

        if(is_nxdomain == 0) {
                /* 0x8180: QR = 1; Opcode = 0; AA = 0; TC = 0; RD = 1; RA = 1;
                 * Z = 0; RCODE = 0 */
                out = make_dns_header(id,0x8180,ancount,nscount,arcount);
        } else if(is_nxdomain == 1) {
                /* Same header as before, but with RCODE of "name error" */
                out = make_dns_header(id,0x8183,ancount,nscount,arcount);
        } else {
                goto catch_make_dns_packet;
        }

        if(dw_append(question,out) == -1 ||
           dw_put_u16(out,1,-1) == -1 || /* QCLASS: 1 */
           dw_append(answer,out) == -1) {
                goto catch_make_dns_packet;
        }

        return out;

catch_make_dns_packet:
        if(out != 0) {
                dw_destroy(out);
                out = 0;
        }
        return 0;
}

#ifndef NOCACHE
/* See if a given reply is in the cache; if so, send them the reply
 * from the cache */
int get_reply_from_cache(dw_str *query, sockaddr_all_T *client,
                         SOCKET sock, int32_t id, int resurrect) {
        dw_str *value = 0; /* Element in cache */
        dw_str *packet = 0;
        socklen_t c_len = sizeof(struct sockaddr_in);
	int ret = -1;

#ifdef IPV6
       if (client->Family == AF_INET6)
               c_len = sizeof(struct sockaddr_in6);
#endif

        value = dwh_get(cache,query,resurrect);
        if(value == 0) {
                return 0; /* Not in cache */
        } 

	if(value->len == 7) { /* Empty packet; workaround */
		dw_log_string("Warning: Removing empty packet from cache",11);
		dwh_zap(cache,query,0);
		return 0; /* Not in cache anymore */
	}

        packet = make_dns_packet(query,value,id);
        if(packet == 0) {
                goto catch_get_reply_from_cache;
        }

        sendto(sock,(void *)packet->str,packet->len,0,
               (struct sockaddr *)client, c_len);

        ret = 1;

catch_get_reply_from_cache:
        if(value != 0) {
                dw_destroy(value);
                value = 0;
        }
        if(packet != 0) {
                dw_destroy(packet);
                packet = 0;
        }
        return ret;
}
#endif /* NOCACHE */

/* Given a connection we will send on, try and send the connection on.
   If we're unable to send the connection on, see if we have an
   expired element with the data we want. */
void try_forward_local_udp_packet(SOCKET sock, int32_t local_id,
     ip_addr_T *from_ip, uint16_t from_port, unsigned char *packet,
     int len, sockaddr_all_T *client,dw_str *query) {

        /* If not cached, get a reply that we will cache and send back to
         * the client */
        if(forward_local_udp_packet(sock,local_id,from_ip,from_port,
                                    packet,len) != -1) {
		return; /* Success! */
	}

#ifndef NOCACHE
	/* OK, at this point it failed so we'll see if we get a
         * "resurrected" cache entry */
        if(resurrections == 1 &&
           get_reply_from_cache(query,client,sock,local_id,1) == 1) {
		dw_log_string("Resurrected from cache",11); 
		return; /* Resurrected entry; we're done */
        }
#endif /* NOCACHE */

        if(handle_overload == 1) {
                send_server_fail(client,packet,len,sock);
        }
}

/* Get and process a local DNS request */
void get_local_udp_packet(SOCKET sock) {
        unsigned char packet[522];
        int len = 0;
        sockaddr_all_T client;
        socklen_t c_len = 0;
        ip_addr_T from_ip = {0,{0,0}};
        uint16_t from_port = 0;
        int32_t local_id = -1;
        dw_str *query = 0;
#ifdef VALGRIND_NOERRORS
	memset(packet,0,522);
#endif /* VALGRIND_NOERRORS */

        c_len = sizeof(client);
        make_socket_nonblock(sock); /* Linux bug workaround */
        len = recvfrom(sock,(void *)packet,520,0,(struct sockaddr *)&client,
                       &c_len);

        from_port = get_from_ip_port(&from_ip,&client);

        if(check_ip_acl(&from_ip) != 1) { /* Drop unauthorized packets */
                goto catch_get_local_udp_packet;
        }

        local_id = get_dns_qid(packet, len, 2);
        if(local_id < 0 || len < 13) { /* Invalid remote packet */
                goto catch_get_local_udp_packet;
        }

        /* See if we have something in the cache; destroyed at end of
         * function */
        query = dw_get_dname_type(packet,12,len);
        if(query == 0) {
                goto catch_get_local_udp_packet;
        }
#ifndef NOCACHE
        if(get_reply_from_cache(query,&client,sock,local_id,0) == 1) {
                goto catch_get_local_udp_packet; /* In cache; we're done */
        }
#endif /* NOCACHE */

	/* Nothing in cache; lets try sending the packet upstream */
        try_forward_local_udp_packet(sock,local_id,&from_ip,from_port,
                                     packet,len,&client,query);

catch_get_local_udp_packet:
        if(query != 0) {
                dw_destroy(query);
                query = 0;
        }
}

/* Forward a remote reply back to the client */
void forward_remote_reply(unsigned char *packet, size_t len, remote_T *r_ip) {
        sockaddr_all_T client;
        socklen_t len_inet = 0;

        if(r_ip == 0) {
                return;
        }
        if(r_ip->from_socket == -1) {
                return;
        }
        memset(&client,0,sizeof(client));
        len_inet = sizeof(client);

        if (r_ip->ip.len == 4) {
                client.V4.sin_family = AF_INET;
                client.V4.sin_port = htons(r_ip->port);
                memcpy(&client.V4.sin_addr, r_ip->ip.ip, 4);
                len_inet = sizeof(struct sockaddr_in);
#ifdef IPV6
        } else if(r_ip->ip.len == 16) {
                client.V6.sin6_family = AF_INET6;
                client.V6.sin6_port = htons(r_ip->port);
                memcpy(&client.V6.sin6_addr, r_ip->ip.ip, 16);
                len_inet = sizeof(struct sockaddr_in6);
#endif
        } else {
                return;
        }
        sendto(r_ip->from_socket,(void *)packet,len,0,
               (struct sockaddr *)&client,len_inet);
}

#ifndef NOCACHE
/* Add a reply we have received from the remote (upstream) DNS server to
 * the cache */
int cache_dns_reply(unsigned char *packet, int count) {
        int32_t ttl = 60;
	int32_t ancount = 0;
        int is_nxdomain = 0;
        dw_str *question = 0, *answer = 0;
	int ret = -1;

        question = dw_get_dname_type(packet,12,count);
        if((packet[3] & 0x0f) == 3) {
                is_nxdomain = 1;
        }
        answer = dw_packet_to_cache(packet,count,is_nxdomain);
	ancount = dw_cachepacket_to_ancount(answer);
	if(ancount == 0) { 
		ancount = 32; /* So we can correctly cache negative answers */
	}

        if(question == 0 || answer == 0 || ancount == -1) {
                goto catch_cache_dns_reply;
        }

        ttl = dw_get_a_dnsttl(answer,0,max_ttl,ancount);

        if(ttl == -1) {
                goto catch_cache_dns_reply;
        }

        /* Physically put the data in the cache */
        dwh_add(cache,question,answer,ttl);
	ret = 1;

catch_cache_dns_reply:
        if(question != 0) {
                dw_destroy(question);
                question = 0;
        }
        if(answer != 0) {
                dw_destroy(answer);
                answer = 0;
        }
	return ret;
}
#endif /* NOCACHE */

/* Verify that a given DNS packet is good (The Query ID is correct, the
   query in the "question" section of the DNS header is good) */
int verify_dns_packet(int b, unsigned char *packet, int len) {
	int ret = 0;
	dw_str *question = 0;
	
        /* Make sure the ID we got is the same as the one we originally
         * sent them */
        if(get_dns_qid(packet,len,0) != rem[b].remote_id) {
                goto catch_verify_dns_packet;
        }

	question = dw_get_dname_type(packet,12,len);
	if(question == 0) {
		goto catch_verify_dns_packet;
	}

	/* Should we make this case-insensitive? Probably not. */
	if(dw_issame(question,rem[b].query) != 1) {
		goto catch_verify_dns_packet;
	}

	ret = 1;	

catch_verify_dns_packet:
	if(question != 0) {
		dw_destroy(question);
		question = 0;
	}	
	return ret;
}

/* Get and process a remote DNS request */
void get_remote_udp_packet(int b, SOCKET sock) {
        ssize_t count;
        unsigned char a[520];

        count = recv(sock,a,514,0);
        if(count < 12 || count > 512) {
                return;
        }
        a[2] |= 0x80; /* Flag this as an answer (just in case they didn't) */

	/* Make reasonably sure this is the reply to their question */
	if(verify_dns_packet(b,a,count) != 1) {
		return;
	}

        if((a[2] & 0x02) == 0x00) { /* If not truncated */
		fflush(stdout);
#ifndef NOCACHE
                if((cache_dns_reply(a,count) == -1) && (deliver_all == 0)) {
			return; /* Bad reply */
		}
#endif /* NOCACHE */
        }

        /* Now make sure the ID is the same as the one the client
         * originally sent us */
        set_dns_qid(a,count,rem[b].local_id);

        /* Send the answer we got to the appropriate local connection */
        forward_remote_reply(a,count,&rem[b]);

        /* Reset this pending remote connection */
        closesocket(b_remote[b]);
        b_remote[b] = -1;
        reset_rem(b);
}

/* Send a server failure back to the client when there is no reply from
 * the upstream server.  Input: The pending remote connection number. */
void server_fail_noreply(int a) {
	dw_str *packet = 0;

	/* 0x8182: QR = 1; Opcode = 0; AA = 0; TC = 0; RD = 1; RA = 1;
         *         Z = 0; RCODE = "server fail" (2) */
	packet = make_dns_header(rem[a].local_id,0x8182,0,0,0);

	if(dw_append(rem[a].query,packet) == -1 ||
	   dw_put_u16(packet,1,-1) == -1 /* QCLASS: 1 */) {
		goto catch_server_fail_noreply;
	}

	forward_remote_reply((unsigned char *)packet->str,packet->len,&rem[a]);
	
catch_server_fail_noreply:
	if(packet != 0) {
		dw_destroy(packet);
		packet = 0;
	}
}

